@@ -2,11 +2,7 @@ import os
 import yaml
 import warnings
 import joblib
-import numpy as np
 import pandas as pd
-import importlib
-import operator
-import warnings
 
 from . import helpers
 from . import feature_helpers
@@ -22,45 +18,30 @@ class Dataset(object):
 
     Attributes
     ----------
-        settings: dict
-            the required keys for the dictionary:
-                'type': which subclass the settings belong to (required for the high-level API only and needs to follow a specific format),
-                'file_name': pickled filename (not the full path) to read the data from, or dump the data to,
-                'curves': list of features from the raw data to use,
-                'id_column': name of the id column,
-                'label_column': name of the column containing the labels,
-                'num_filler': filler value for numerical curves(existing or wishing value for replacing missing values),
-                'cat_filler': filler value categorical curves(existing or wishing value for replacing missing values),
-                'scaler_method': (options: 'StandardScaler', 'RobustScaler'),
-                'gradient_features': whether to calculate gradients of features (boolean),
-                'window_features': whether to calculate rolling aggregations (min, max, mean) of features (boolean),
-                'window': size of the rolling window for calculating the window features (int),
-                'log_features': whether to add log of every feature as a new feature (boolean)
-        *all setting keys are also set as class attributes
+    
+        settings: dict/path to a yaml file
+            the possible keys for the settings:
+                "file_name" (required): pickled filename (not the full path) to read the data from, or dump the data to,
+                "curves" (required): list of features from the raw data to use,
+                "id_column" (required): name of the id column, eg. well_name
+                "label_column" (required): name of the column containing the labels
+                "curves_to_scale" (optional - default []): list of curves from the curves list should be scaled,
+                "curves_to_normalize" (optional - default []): list of curves from the curves list should be 2-point normalized
+                "num_filler" (optional - default 0): filler value for numerical curves(existing or wishing value for replacing missing values)
+                "cat_filler" (optional - default 'MISSING'): filler value categorical curves(existing or wishing value for replacing missing values)
+                "scaler_method" (optional - default 'RobustScaler'): (options: 'StandardScaler', 'RobustScaler', 'MinMaxScaler')
+                "gradient_features" (optional - default []): list of features from curves to calculate gradients of
+                "window_features" (optional - default []): list of features from curves to calculate rolling aggregations (min, max, mean) of
+                "window" (optional - default 1): size of the rolling window for calculating the window features (int)
+                "log_features" (optional - default []): list of features from curves to calculate log
+                "drop_original_curves" (optional - default False): boolean, whether to keep or not the original curves (that were converted to window, lof or gradient)
+                "noise_removal_window" (optional - default None): int, if a median filtering of each curve is required. if None no filtering is applied                "imputer": IterativeImputer or SimpleImputer
+                "categorical_curves" (optional - default []): list of curves which are categorical. If none is specified the algorithm tries to determine the categorical columns
+                "feat_eng" (optional - default None): dictionary where keys are curves to generate and values 0/1 for not scaling or scaling. if None no feature engineering is applied
+        **all setting keys are also set as class attributes**
 
         folder_path: the path to the local copy of the data as well as the serialized scaler and imputer(if applicable)
 
-    Methods
-    -------
-        load_from_cdf(client:CogniteClient, reload:bool)
-            - loads data from CDF and dumps it in folder_path/file_name if reload=True
-            - loads data from folder_path/file_name if reload=False
-            - populates the 'df_original' attribute
-
-        load_from_csv(filepath:str)
-            - loads local data from filepath into the 'df_original' attribute
-
-        load_from_pickle(filepath:str)
-            - loads local data from filepath into the 'df_original' attribute
-
-        load_from_dict(filepath:str)
-            - loads local data from filepath into the 'df_original' attribute
-
-        preprocess(train:bool)
-            - performs all preprocessing according to the settings
-            - splits data into train and test set (if train=True)
-            - saves scaler and imputer to folder_path/scaler.joblib and folder_path/imputer.joblib (if applicable)
-                or loads them from the mentioned paths to use for preprocessing
     """
 
     def __init__(self, mappings, settings, folder_path):
@@ -95,6 +76,9 @@ class Dataset(object):
         if not hasattr(self, 'curves_to_scale'):
             setattr(self, 'curves_to_scale', [])
 
+        if not hasattr(self, 'categorical_curves'):
+            setattr(self, 'categorical_curves', [])
+
         if not hasattr(self, 'lognames'):
             setattr(self, 'lognames', [])
         
@@ -116,9 +100,6 @@ class Dataset(object):
         if not hasattr(self, 'drop_original_curves'):
             self.drop_original_curves = False
 
-        if not hasattr(self, 'scaler_method'):
-            self.scaler_method = 'RobustScaler'
-
         if not hasattr(self, 'num_filler'):
             self.num_filler = 0
 
@@ -130,36 +111,40 @@ class Dataset(object):
             setattr(self, 'curves_to_normalize', [])
 
         # Standardize the provided curve names
-        for attr in ['curves', 'curves_to_scale', 'curves_to_normalize', 'gradient_features', 'log_features']:
+        original_curve_names = {}
+        for attr in ['curves', 'curves_to_scale', 'curves_to_normalize', 'gradient_features', 'log_features', 'categorical_curves']:
             setattr(self, '{}_original'.format(attr), getattr(self, attr))
-            setattr(self, attr, helpers._standardize_names(names=getattr(self, attr), mapper=self.curve_mappings))
+            new_names, original_curve_names[attr] =\
+                helpers._standardize_names(names=getattr(self, attr), mapper=self.curve_mappings)
+            setattr(self, attr, new_names)
 
+        # standardize label columns
+        tmp_label_column, original_label_col =\
+            helpers._standardize_names([self.label_column], mapper=self.curve_mappings)
+        self.label_column = tmp_label_column[0]
         # Check that the label curve is not present in the curves
         if self.label_column in self.curves:
-            raise ValueError (f'Label column ({self.label_column}) is present in the input curves.')
+            original_col_label = original_label_col[self.label_column]
+            original_col_curve = original_curve_names['curves'][self.label_column]
+            raise ValueError (f'Label column ({original_col_label}) is present in the input curves ({original_col_curve}).')
 
-        # Check that curves_to_scale is a  subset of curves
-        for attr in ['curves_to_scale']:
+        # Check that curves_to_scale is a subset of curves
+        for attr in ['curves_to_scale', 'curves_to_normalize', 'gradient_features', 'log_features', 'categorical_curves']:
             all_curves = set(getattr(self, 'curves'))
             sub_curves = set(getattr(self, attr))
-            if not sub_curves.issubset(all_curves): 
-                raise ValueError ('{} is not a subset of curves {}'.format(attr, getattr(self, 'curves')))
+            if not sub_curves.issubset(all_curves):
+                sub_curves = [original_curve_names[attr][c] for c in sub_curves]
+                all_curves = [original_curve_names['curves'][c] for c in all_curves]
+                raise ValueError ('{} is not a subset of curves {}'.format(sub_curves, all_curves))
 
         if hasattr(self, 'imputer'):
             self.imputer_path = os.path.join(self.folder_path, 'imputer.joblib')
         else:
             self.imputer = None
 
-        self.scaler_path = os.path.join(self.folder_path, 'scaler.joblib')
         self.data_path   = os.path.join(self.folder_path, self.file_name)
 
-    def _read_data(self):
-        """
-        Identify column type and checks all expected curves are present
-        """
-        self._validate_data()
-
-    def load_from_cdf(self, client:CogniteClient, reload=False):
+    def load_from_cdf(self, client, reload=False):
         """                
         Loads and validates data
         
@@ -168,9 +153,12 @@ class Dataset(object):
             reload: boolean - default False
                 if reload is True: downloads data from CDF as a pandas DataFrame and saves model_path/file_name as in Dataset settings
                 if reload is False: expects pickled file to exist at model_path/file_name, otherwise loads it from CDF
+
+        Returns:
+            None
         """
         self.df_original = self._load_data_faster(client, reload=reload)
-        self._read_data()
+        self._validate_data()
 
     def load_from_csv(self, filepath, **kwargs):
         """
@@ -178,29 +166,39 @@ class Dataset(object):
 
         Args:
             filepath (string): path to csv file
+
+        Returns:
+            None
         """
         self.df_original = pd.read_csv(filepath, **kwargs)
-        self._read_data()
+        self._validate_data()
 
     def load_from_pickle(self, filepath, **kwargs):
         """
-        Loads data from pickle file and validates
+        Loads data from pickle files and validates data
 
-        Args:       
+        Args:
             filepath (string): path to pickle file
+
+        Returns:
+            None
         """
+
         self.df_original = pd.read_pickle(filepath, **kwargs)
-        self._read_data()
+        self._validate_data()
 
     def load_from_dict(self, data_dict):
         """
         Loads data from a dictionary and validates it
 
         Args:
-            data_dict (string): dictionary with data
+            data_dict (dict): dictionary with data
+
+        Returns:
+            None
         """
         self.df_original = pd.DataFrame.from_dict(data_dict)
-        self._read_data()
+        self._validate_data()
 
     def load_from_df(self, data_df):
         """
@@ -208,9 +206,12 @@ class Dataset(object):
 
         Args:
             data_df (pandas.Dataframe): dataframe with data
+
+        Returns:
+            None
         """
         self.df_original = data_df
-        self._read_data()
+        self._validate_data()
 
     def train_test_split(self, test_size=0.2, df=None, **kwargs):
         """
@@ -239,7 +240,7 @@ class Dataset(object):
         Returns:
             tuple: input (features) and output (target) dataframes
         """
-        df = helpers._standardize_curve_names(df, mapper=kwargs.get('curve_mappings', self.curve_mappings))
+        df, _ = helpers._standardize_curve_names(df, mapper=kwargs.get('curve_mappings', self.curve_mappings))
         if self.id_column not in self.curves:
             X = df.loc[:, ~df.columns.isin([self.label_column, self.id_column])]
         else:
@@ -249,25 +250,73 @@ class Dataset(object):
 
     def preprocess(self, df, **kwargs):
         """
-        Main preprocess function, including all required steps by the user
+        Main preprocessing function, includs the following steps (some methods may be excluded following the settings provided at init):
+
+            - standardize curve names - helpers._standardize_curve_names()
+
+            - apply metadata - helpers._apply_metadata()
+
+            - clean data by applying certain physical cut-offs - helpers._remove_cutoff_values()
+
+            - impute values with depth rende method - imputation_helpers._impute_depth_trend()
+
+            - remove noise - _remove_noise()
+
+            - generate specified features - _feature_engineer()
+            
+            - normalize specified curves - _normalize_curves()
+
+            - apply per-well preprocessing and feature engineering - _preprocess()
+
+            - fill the missing Z_LOC values with regards to DEPTH_MD (assumed to always be present)
+
+            - encode categorical features
+
+            - scales specified curves
 
         Args:
             df (pd.Dataframe): dataframe to which apply preprocessing
+            
+        Keyword Args:
 
+            _metadata = {
+                    'num_filler': None,
+                    'cat_filler': None
+                    }
+
+            _depth_trend_imputation = {
+                    'curves': None,
+                    'imputation_models': None,
+                    'save_imputation_models': False,
+                    'allow_individual_models': True
+                    }
+
+            _normalize_curves = {
+                    'low_perc': 0.05,
+                    'high_perc': 0.95,
+                    'key_wells': None,
+                    'save_key_wells': False
+                    }
+
+            _scaling = {
+                    'scaler_method': RobustScaler,
+                    'scaler': None,
+                    'save_scaler': False
+                    }
+                        
         Returns:
             pd.Dataframe: preprocessed dataframe
             dictionary: key wells and values for each curve
             list: feats_to_keep specifies which curves should be kept in accordance to the settings file
         """
-        
-        # identify categorical and numerical curves, impute if needed
-        df, num_cols, cat_cols = helpers._apply_metadata(df, **kwargs.get('_metadata', {}))
         # standardize curve names
-        df = helpers._standardize_curve_names(df, mapper=kwargs.get('curve_mappings', self.curve_mappings))
+        df, old_new_cols = helpers._standardize_curve_names(df, mapper=kwargs.get('curve_mappings', self.curve_mappings))
+        # identify categorical and numerical curves, impute if needed
+        df, num_cols, cat_cols = helpers._apply_metadata(df, old_new_cols, self.categorical_curves, **kwargs.get('_metadata', {}))
         # remove or clip values in some curves
-        df = helpers._remove_cutoff_values(df, self.curves)
+        df = helpers._remove_cutoff_values(df, self.curves, old_new_cols)
         # impute value with depth trend
-        df = imputation_helpers._impute_depth_trend(df, self.folder_path, **kwargs.get('_depth_trend_imputation', {}))
+        df = imputation_helpers._impute_depth_trend(df, self.folder_path, self.curve_mappings, **kwargs.get('_depth_trend_imputation', {}))
         # remove noise if chosen
         if self.noise_removal_window is not None:
             df = self._remove_noise(df, cols=list(num_cols))
@@ -277,7 +326,7 @@ class Dataset(object):
         # normalize some curves 
         df, key_wells = self._normalize_curves(df, **kwargs.get('_normalize_curves', {}))
         # preprocess whole dataset (including add features to the dataset)
-        df, added_cols = self._preprocess(df, **kwargs.get('_preprocess', {}))
+        df, added_cols = self._preprocess(df)
         
         # preprocess some curves per well
         if self.id_column in df.columns:
@@ -286,7 +335,7 @@ class Dataset(object):
             time_cols  = None
             for well in well_names:
                 well_df                            = df.loc[df[self.id_column]==well, :].copy()
-                well_df, new_cols, curves_to_scale = self._process_well(well_df, **kwargs.get('_process_well', {}))
+                well_df, new_cols, curves_to_scale = self._process_well(well_df, old_new_cols)
                 if 'time_features' in kwargs.keys():
                     time_cols   = [c for c in df.columns if c != self.id_column and c != self.label_column]
                     well_df, time_cols = feature_helpers._add_time_features(well_df, time_cols, kwargs['time_features'])
@@ -294,19 +343,20 @@ class Dataset(object):
             df = res_df.copy()
             curves_to_scale = curves_to_scale + time_cols if time_cols!=None else curves_to_scale
         else:
-            df, new_cols, curves_to_scale = self._process_well(df, **kwargs.get('_process_well', {}))
+            df, new_cols, curves_to_scale = self._process_well(df, old_new_cols)
             warnings.warn('Not possible to process per well as well ID is not in dataset. Preprocessing was done considering all data is from the same well.')
         
         # fill the missing Z_LOC values with regards to DEPTH_MD(always present)
         if ("Z_LOC" in self.curves) and ("DEPTH_MD" in self.curves):
             df.loc[:, 'Z_LOC'] = df['Z_LOC'].fillna(-(df['DEPTH_MD'] - 20))
         # impute missing rows
-        num_cols, cat_cols  = helpers._make_col_dtype_lists(df)
+        num_cols, cat_cols  = helpers._make_col_dtype_lists(df, old_new_cols, self.categorical_curves)
         df.loc[:, num_cols] = df[num_cols].fillna(self.num_filler)
         df.loc[:, cat_cols] = df[cat_cols].fillna(self.cat_filler)
 
         # map the categorical features
         df = self._encode_columns(df, columns=cat_cols)
+
         # drop original values if chosen otherwise all columns will be considered
         feats_to_keep = list(df.columns)
         if self.drop_original_curves:
@@ -319,7 +369,6 @@ class Dataset(object):
                 to_drop = to_drop + time_cols_drop
             feats_to_keep = [c for c in df.columns if c not in to_drop]
         
-
         # make sure we only have numerical features to scale
         num_feats       = set(df[feats_to_keep]._get_numeric_data().columns)
         curves_to_scale = list(set(curves_to_scale) & num_feats)
@@ -327,10 +376,9 @@ class Dataset(object):
         if len(curves_to_scale) > 0:
             df.loc[:, curves_to_scale] = self._scale_columns(
                 df, 
-                columns=curves_to_scale, 
-                scaler_method=self.scaler_method, 
-                **kwargs
-            )      
+                columns=curves_to_scale,
+                **kwargs.get('_scaling', {})
+            )
         
         added_cols = added_cols + new_cols
         self.added_features = added_cols
@@ -454,13 +502,15 @@ class Dataset(object):
         """
         Checks that the data loaded into the Dataset includes the expected curves
         """
+        # standardize curve names
+        df_original, _ = helpers._standardize_curve_names(self.df_original, mapper=self.curve_mappings)
         # check that all expected curves are present in the data
         expected_curves = self.curves
-        present_curves  = self.df_original.columns.tolist()
+        present_curves  = df_original.columns.tolist()
         expected_but_missing_curves = [c for c in expected_curves if c not in present_curves]
         if expected_but_missing_curves:
             warnings.warn("Warning...........There are curves that are expected but missing from data. They are being filled with num_filler {}".format(expected_but_missing_curves))
-            self.df_original[expected_but_missing_curves] = self.num_filler
+            df_original[expected_but_missing_curves] = self.num_filler
 
     def _encode_columns(self, df, columns=None):
         """
@@ -520,6 +570,15 @@ class Dataset(object):
         Args:
             df (pd.DataFrame): dataframe with columns to normalize
 
+        Keyword Args:
+            low_perc (float): low quantile to use as min value
+
+            high_perc (float): high quantile to use as max value
+            
+            user_key_wells (dict): dictionary with curves as keys and min/max values and key well as values
+            
+            save_key_wells (bool): whether to save keys wells dictionary in folder_path
+        
         Returns:
             tuple: pd.DataFrame with normalized values and dictionary with key wells and values
         """
@@ -528,16 +587,14 @@ class Dataset(object):
         user_key_wells = kwargs.get('key_wells')
         save_key_wells = kwargs.get('save_key_wells', False)
         
+        wells_data = df.groupby(self.id_column)
         if user_key_wells is None:
-            wells_data = df.groupby(self.id_column)
             key_wells  = {c: None for c in self.curves_to_normalize}
             for c in self.curves_to_normalize:
                 low_p  = wells_data[c].quantile(low_perc)
                 high_p = wells_data[c].quantile(high_perc)
-
                 # get the key well with largest difference between limit values
                 key_well = (high_p-low_p).idxmax()
-
                 df.loc[:, 'low_p']  = df[self.id_column].map(low_p)
                 df.loc[:, 'high_p'] = df[self.id_column].map(high_p)
                 # save the key wells for normalization
@@ -547,49 +604,39 @@ class Dataset(object):
                     'ref_low': df[df[self.id_column]==key_well]['low_p'].unique()[0],
                     'ref_high': df[df[self.id_column]==key_well]['high_p'].unique()[0]
                 }
-                # normalize all other wells
-                df.loc[:, c] = df.apply(
-                    lambda x: helpers._normalize(
-                        x[c],
-                        key_wells[c]['ref_low'],
-                        key_wells[c]['ref_high'],
-                        x['low_p'],
-                        x['high_p'],
-                        ),
-                        axis=1
-                        )
         else:
             # if key wells is provided as a dict with the same format
-            if isinstance(user_key_wells, dict):
-                if user_key_wells.keys() != set(self.curves_to_normalize):
-                    raise ValueError(
-                        "Curves included in the key wells dictionary incosistent with curves_to_normalize", 
-                        user_key_wells.key(), self.curves_to_normalize)
-                else:
-                    key_wells = user_key_wells
-                    for c in self.curves_to_normalize:
-                        df.loc[:, 'low_p']  = key_wells[c]['ref_low']
-                        df.loc[:, 'high_p'] = key_wells[c]['ref_high']
-                        # normalize all other wells
-                        df.loc[:, c] = df.apply(
-                            lambda x: helpers._normalize(
-                                x[c],
-                                key_wells[c]['ref_low'],
-                                key_wells[c]['ref_high'],
-                                x['low_p'],
-                                x['high_p'],
-                                ),
-                                axis=1
-                                )
-            else:
-                ValueError("Other methods to provide key wells are not implemented yet!") 
-        
+            if not isinstance(user_key_wells, dict):
+                raise ValueError("Other methods to provide key wells are not implemented yet!") 
+            if user_key_wells.keys() != set(self.curves_to_normalize):
+                raise ValueError(
+                    "Curves included in the key wells dictionary inconsistent with curves_to_normalize", 
+                    user_key_wells.keys(),
+                    self.curves_to_normalize
+                )
+            key_wells = user_key_wells
+        # normalize all wells
+        for c in self.curves_to_normalize:
+            low_p  = wells_data[c].quantile(low_perc)
+            high_p = wells_data[c].quantile(high_perc)
+            df.loc[:, 'low_p']  = df[self.id_column].map(low_p)
+            df.loc[:, 'high_p'] = df[self.id_column].map(high_p)
+            df.loc[:, c] = df.apply(
+                lambda x: helpers._normalize(
+                    x[c],
+                    key_wells[c]['ref_low'],
+                    key_wells[c]['ref_high'],
+                    x['low_p'],
+                    x['high_p'],
+                ),
+                axis=1
+            )
         if save_key_wells:                          
             # save key wells to where model is
             joblib.dump(key_wells, os.path.join(self.folder_path, 'key_wells.joblib'))
         return df, key_wells
 
-    def _preprocess(self, df, **kwargs):
+    def _preprocess(self, df):
         """
         Preprocessing pipeline for all wells, independently of well ID
 
@@ -609,7 +656,7 @@ class Dataset(object):
             added_cols   = added_cols + log_cols
         return df, added_cols
 
-    def _scale_columns(self, df, columns, scaler_method='RobustScaler', **kwargs):
+    def _scale_columns(self, df, columns, **kwargs):
         """
         Scales specified columns
 
@@ -618,22 +665,34 @@ class Dataset(object):
             columns (list): list with columns to scale
             scaler_method (str, optional): scaling method. Defaults to 'RobustScaler'.
 
+        Keyword Args:
+            scaler_method (str): string of any sklearn scalers
+
+            scaler (sklearn.preprocessing): scaler object from sklearn
+            
+            save_scaler (bool): whether to save scaler in folder_path or not
+
         Returns:
             pd.DataFrame: scaled columns
         """
-        cls_name = scaler_method
-        try:
-            cls = getattr(sklearn.preprocessing, cls_name)
-        except AttributeError as ae:
-            print(ae)
-        self.scaler = cls(**kwargs.get('scaler', {}))
-        self.scaler.fit(df[columns])
-        # save scaler to same path as model
-        scaler_path = os.path.join(self.folder_path, 'scaler.joblib')
-        joblib.dump(self.scaler, scaler_path)
-        return self.scaler.transform(df[columns])
+        scaler_method = kwargs.get('scaler_method', 'RobustScaler')
+        scaler        = kwargs.get('scaler', None)
+        save_scaler   = kwargs.get('save_scaler', False)
 
-    def _process_well(self, df, **kwargs):
+        if scaler is None:
+            try:
+                scaler = getattr(sklearn.preprocessing, scaler_method)
+            except AttributeError as ae:
+                print(ae)
+            scaler = scaler(**kwargs.get('scaler', {}))
+            scaler.fit(df[columns])
+            # save scaler to same path as model
+            if save_scaler:
+                scaler_path = os.path.join(self.folder_path, 'scaler.joblib')
+                joblib.dump(scaler, scaler_path)
+        return scaler.transform(df[columns])
+
+    def _process_well(self, df, old_new_cols):
         """
         Process specific well: imputation, features creation (rolling, gradient)
 
@@ -645,14 +704,14 @@ class Dataset(object):
         """
         # impute features
         if self.imputer == 'iterative':
-            df = imputation_helpers._iterative_impute(df)
+            df = imputation_helpers._iterative_impute(df, self.categorical_curves, old_new_cols)
         if self.imputer == 'simple':
-            df = imputation_helpers._simple_impute(df)
+            df = imputation_helpers._simple_impute(df, self.categorical_curves, old_new_cols)
 
         added_cols = []
         # add rolling features
         if self.window_features:
-            rolling_columns, _ = helpers._make_col_dtype_lists(df[self.curves])
+            rolling_columns, _ = helpers._make_col_dtype_lists(df[self.curves], old_new_cols, self.categorical_curves)
             df.loc[:, self.curves], window_cols = feature_helpers._add_rolling_features(df, rolling_columns, self.window)
             added_cols = added_cols + window_cols
 
