@@ -20,15 +20,6 @@ def _standardize_group_formation_name(name):
         name=name.replace('Ø','O')
     return name
 
-def generate_log_settings(settings, lognames, badlognames, log_curves=None):
-    all_settings = dict()
-    for i, logname in enumerate(lognames):
-        all_settings[logname] = settings
-        all_settings[logname]['label_column'] = badlognames[i]
-        if log_curves is not None:
-            all_settings[logname]['curves'] = log_curves
-    return all_settings
-
 def _standardize_names(names, mapper):
     """
     Standardize curve names in a list based on the curve_mappings dictionary. 
@@ -47,7 +38,8 @@ def _standardize_names(names, mapper):
             standardized_names.append(mapper.get(name))
         else:
             standardized_names.append(name)
-    return standardized_names
+    old_new_cols = {n: o for o, n in zip(names, standardized_names)}
+    return standardized_names, old_new_cols
 
 def _standardize_curve_names(df, mapper):
     """
@@ -61,9 +53,11 @@ def _standardize_curve_names(df, mapper):
     Returns:
         pd.DataFrame: dataframe with columns names standardized
     """
-    return df.rename(columns=mapper)
+    new_cols_df = df.rename(columns=mapper)
+    old_new_cols = {n: o for o, n in zip(df.columns, new_cols_df.columns)}
+    return new_cols_df, old_new_cols
 
-def _make_col_dtype_lists(df):
+def _make_col_dtype_lists(df, old_new_mapper, cat_original=[]):
     """
     Returns lists of numerical and categorical columns
 
@@ -73,9 +67,22 @@ def _make_col_dtype_lists(df):
     Returns:
         tuple: lists of numerical and categorical columns
     """
+    cat_original = set(cat_original)
     num_cols = set(df._get_numeric_data().columns)
-    cat_cols = list(set(df.columns) - set(num_cols))
-    return list(num_cols), cat_cols
+    cat_cols = set(df.columns) - num_cols
+    if cat_cols != cat_original:
+        extra = cat_original-cat_cols
+        if len(extra):
+            extra = [old_new_mapper[c] for c in extra]
+            warnings.warn(f"Cols {extra} were specified as categorical by user eventhough are numerical.")
+        extra = cat_cols - cat_original
+        if len(extra):
+            extra = [old_new_mapper[c] for c in extra]
+            warnings.warn(f"Cols {extra} were identified as categorical and are being added to settings categorical curves.")
+    cat_cols = cat_original.union(cat_cols)
+    # make sure nothing from categorical is in num cols
+    num_cols = num_cols-cat_cols
+    return list(num_cols), list(cat_cols)
 
 def generate_sequential_dataset(df, time_steps):
     """
@@ -121,41 +128,61 @@ def oversample_targets(df, label_column, method="RandomOverSampler"):
     Returns:
         pd.DataFrame: oversampled dataframe
     """
-    features                    = df.loc[:, df.columns != label_column]
-    targets                     = df.loc[:, label_column]
-    module_name                 = "imblearn.over_sampling"
-    cls_name                    = method
-    cls                         = getattr(importlib.import_module(module_name), cls_name)
-    oversampler                 = cls()
-    features, targets           = oversampler.fit_resample(features, targets)
+    features               = df.loc[:, df.columns != label_column]
+    targets                = df.loc[:, label_column]
+    module_name            = "imblearn.over_sampling"
+    cls_name               = method
+    cls                    = getattr(importlib.import_module(module_name), cls_name)
+    oversampler            = cls()
+    features, targets      = oversampler.fit_resample(features, targets)
     features[label_column] = targets
     return features
 
-def _apply_metadata(df, **kwargs):
+def undersample_targets(df, label_column, method="RandomUnderSampler"):
+    """
+    Undersamples dataset
+
+    Args:
+        df (pd.DataFrame): dataframe to undersample
+        method (str, optional): method used for undersampling. Defaults to "RandomUnderSampler".
+
+    Returns:
+        pd.DataFrame: undersampled dataframe
+    """
+    features               = df.loc[:, df.columns != label_column]
+    targets                = df.loc[:, label_column]
+    module_name            = "imblearn.under_sampling"
+    cls_name               = method
+    cls                    = getattr(importlib.import_module(module_name), cls_name)
+    undersampler            = cls()
+    features, targets      = undersampler.fit_resample(features, targets)
+    features[label_column] = targets
+    return features
+
+def _apply_metadata(df, old_new_cols, cat_curves=[], **kwargs):
     """
     Applies specified metadata to data
 
     Args:
         df (pd.DataFrame): dataframe to apply metadata to
 
+    Keyword Args:
+        num_filler (float/int): which filler for numerical data has been used for imputation, if any
+        cat_filler (str): which filler for categorical data has been used for imputation, if any
+
     Returns:
         tuple: pd.Dataframe after applying metadata, list of numerical columns and list of categorical columns
     """
-    num_cols, cat_cols = _make_col_dtype_lists(df)
-    imputed            = kwargs.get('imputed', False)
-    if imputed:
-        print('Applying metadata....')
-        if 'num_filler' in kwargs.keys():
-            num_filler = kwargs['num_filler']
-            if imputed:
-                df.loc[:, num_cols] = df[num_cols].replace(to_replace=num_filler, value=np.nan)
-        if 'cat_filler' in kwargs.keys():
-            cat_filler = kwargs['cat_filler']
-            if imputed:
-                df.loc[:, cat_cols] = df[cat_cols].replace(to_replace=cat_filler, value=np.nan)
+    num_cols, cat_cols = _make_col_dtype_lists(df, old_new_cols, cat_curves)
+    num_filler         = kwargs.get('num_filler', None)
+    cat_filler         = kwargs.get('cat_filler', None)
+    if num_filler is not None:
+        df.loc[:, num_cols] = df[num_cols].replace(to_replace=num_filler, value=np.nan)
+    if cat_filler is not None:
+        df.loc[:, cat_cols] = df[cat_cols].replace(to_replace=cat_filler, value=np.nan)
     return df, num_cols, cat_cols 
 
-def _remove_cutoff_values(df, curves, th=0.05):
+def _remove_cutoff_values(df, curves, old_new_cols, th=0.05):
     """
     Returns the dataframe after applying the cutoff for some curves
 
@@ -173,7 +200,8 @@ def _remove_cutoff_values(df, curves, th=0.05):
         outliers_low  = df[df.GR<0]
         outliers_high = df[df.GR>250]
         if (len(outliers_low)+len(outliers_high))/len_df>th:
-            warnings.warn("Warning...........GR has more than 5% of its values lower than 0 or higher than 250")
+            old_col = old_new_cols['GR']
+            warnings.warn(f"Warning...........{old_col} has more than 5% of its values lower than 0 or higher than 250")
         df.loc[df.GR<0, 'GR']   = 0
         df.loc[df.GR>250, 'GR'] = 250
     
@@ -181,21 +209,24 @@ def _remove_cutoff_values(df, curves, th=0.05):
         if resistivity in curves:
             outliers = df[df[resistivity]>100]
             if len(outliers)/len_df>th:
-                warnings.warn("Warning...........{} has more than 5% of its values higher than 100")
+                old_col = old_new_cols[resistivity]
+                warnings.warn(f"Warning...........{old_col} has more than 5% of its values higher than 100")
             df.loc[df[resistivity]>100, resistivity] = 100
 
     if 'NEU' in curves:
         outliers_high = df[df.NEU>1]
         outliers_low  = df[df.NEU<-0.5]
         if (len(outliers_low)+len(outliers_high))/len_df>th:
-            warnings.warn("Warning...........NEU has more than 5% of its values higher than 1")
+            old_col = old_new_cols['NEU']
+            warnings.warn(f"Warning...........{old_col} has more than 5% of its values higher than 1")
         df.loc[df.NEU>1, 'NEU']    = np.nan
         df.loc[df.NEU<-0.5, 'NEU'] = np.nan
     
     if 'PEF' in curves:
         outliers = df[df.PEF>10]
         if len(outliers)/len_df>th:
-            warnings.warn("Warning...........PEF has more than 5% of its values higher than 10")
+            old_col = old_new_cols['PEF']
+            warnings.warn(f"Warning...........{old_col} has more than 5% of its values higher than 10")
         df.loc[df.PEF>10, 'PEF'] = np.nan
 
     return df
@@ -237,7 +268,6 @@ def _df_split_train_test(df, id_column, test_size=0.2, test_wells=None):
     return df_train, df_test, test_wells
 
 def _normalize(col, ref_low, ref_high, well_low, well_high):
-
     """
     Formula to normalize min and max values to the key well's min and max values.
 
