@@ -10,7 +10,7 @@ from sklearn.impute import IterativeImputer, SimpleImputer
 from sklearn.linear_model import BayesianRidge, LinearRegression
 from sklearn.preprocessing import PolynomialFeatures
 
-def _simple_impute(df):
+def _simple_impute(df, cat_original=[], old_new_cols={}):
     """
     Imputes missing values in specified columns with simple imputer
 
@@ -20,7 +20,7 @@ def _simple_impute(df):
     Returns:
         pd.DataFrame: dataframe with imputed values
     """
-    num_cols, cat_cols = helpers._make_col_dtype_lists(df)
+    num_cols, cat_cols = helpers._make_col_dtype_lists(df, old_new_cols, cat_original)
 
     # Impute numerical columns
     missing_fractions_num          = df[num_cols].isnull().sum()/df.shape[0]
@@ -46,7 +46,7 @@ def _simple_impute(df):
 
     return df
 
-def _iterative_impute(df, imputer=None):
+def _iterative_impute(df, cat_original=[], old_new_cols={}, imputer=None):
     """
     Imputes missing values in specified columns with iterative imputer
 
@@ -57,7 +57,7 @@ def _iterative_impute(df, imputer=None):
     Returns:
         pd.DataFrame: dataframe with imputed values
     """
-    num_cols, _ = helpers._make_col_dtype_lists(df)
+    num_cols, _ = helpers._make_col_dtype_lists(df, old_new_cols, cat_original)
     # Iterative imputer at 
     missing_fractions = df[num_cols].isnull().sum()/df.shape[0]
     partially_missing_features = missing_fractions.loc[missing_fractions.values!=1].index.tolist()
@@ -157,26 +157,34 @@ def _apply_depth_trend_imputation(df, curves, imputation_models):
             df.loc[missing, c] = poly_preds  
     return df
 
-def _impute_depth_trend(df, folder_path, **kwargs):
+def _impute_depth_trend(df, folder_path, mapper, **kwargs):
     """
     Imputation of curves based on polynomial regression models of the curve based on DEPTH
 
     Args:
         df (pd.DataFrame): df to impute curves
 
-    Raises:
-        ValueError: [description]
+    Keyword Args:
+        curves (list): llist of curves to depth impute
+        imputation_models (dict): dictionary with curves as keys and the sklearn model as value
+        save_imputation_models (bool): whether to sabe the models in the folder_path
+        allow_individual_models (bool): whether to allow individual models if seen that it has enough data
+        to do so (better performance per well)
 
     Returns:
         pd.DataFrame: dataframe with curves imputed
     """
+    
     curves                  = kwargs.get('curves', None)
     imputation_models       = kwargs.get('imputation_models', None)
     save_imputation_models  = kwargs.get('save_imputation_models', False)
-    allow_individual_models = kwargs.get('allow_individual_models', True)
+    allow_individual_models = kwargs.get('allow_individual_models', True)  
 
     if curves is not None:
 
+        # we need to first standardize names
+        curves, _ = helpers._standardize_names(curves, mapper)
+
         # check if depth and all other curves in df
         if not all(c in df.columns for c in curves+['DEPTH']):
             ValueError('Cannot perform depth trend imputation as not all curves are in the dataset.')
@@ -197,10 +205,10 @@ def _impute_depth_trend(df, folder_path, **kwargs):
                     if allow_individual_models:
                         warnings.warn("Some provided curves for imputing do not have models. Models will be generated.")
                     else:               
-                            raise ValueError(
-                                "Curves included in the imputation models dictionary incosistent with curves to impute",
-                                imputation_models.keys(), curves
-                            )
+                        raise ValueError(
+                            "Curves included in the imputation models dictionary inconsistent with curves to impute",
+                            imputation_models.keys(), curves
+                        )
             # check if it is preferable to use individual models instead of given global models
             if allow_individual_models:
                 imputation_models = _individual_imputation_models(df, curves, imputation_models)
